@@ -50,6 +50,7 @@ class Logger:
         slskmessages.DistribEmbeddedMessage,
         slskmessages.DistribSearch,
         slskmessages.EmbeddedMessage,
+        slskmessages.PeerMessageProgress,
         slskmessages.SetConnectionStats,
         slskmessages.SharedFileList,
         slskmessages.UnknownPeerMessage
@@ -100,10 +101,13 @@ class Logger:
             log_file.last_active = time.time()
 
         except Exception as error:
+            # Avoid infinite recursion
+            should_log_file = (folder_path != config.sections["logging"]["debuglogsdir"])
+
             self.add(_("Couldn't write to log file \"%(filename)s\": %(error)s"), {
                 "filename": os.path.join(folder_path, base_name),
                 "error": error
-            })
+            }, should_log_file=should_log_file)
 
     def close_log_file(self, log_file):
 
@@ -155,7 +159,7 @@ class Logger:
 
         return msg
 
-    def add(self, msg, msg_args=None, level=None):
+    def add(self, msg, msg_args=None, level=None, should_log_file=True):
 
         levels = self.log_levels if self.log_levels else config.sections["logging"].get("debugmodes", [])
 
@@ -170,14 +174,16 @@ class Logger:
             if msg_class in self.EXCLUDED_MSGS:
                 return
 
-            msg = "%s %s" % (msg_class, self.contents(msg))
+            msg_direction = "OUT" if msg_args else "IN"
+            msg = "%s: %s %s" % (msg_direction, msg_class, self.contents(msg))
+            msg_args = None
 
         msg = self.set_msg_prefix(level, msg)
 
         if msg_args:
             msg = msg % msg_args
 
-        if config.sections["logging"].get("debug_file_output", False):
+        if should_log_file and config.sections["logging"].get("debug_file_output", False):
             self.write_log_file(
                 folder_path=config.sections["logging"]["debuglogsdir"], base_name=self.file_name, text=msg)
 
@@ -209,8 +215,8 @@ class Logger:
     def add_conn(self, msg, msg_args=None):
         self.add(msg, msg_args=msg_args, level="connection")
 
-    def add_msg_contents(self, msg):
-        self.add(msg, msg_args=None, level="message")
+    def add_msg_contents(self, msg, is_outgoing=False):
+        self.add(msg, msg_args=is_outgoing, level="message")
 
     def add_transfer(self, msg, msg_args=None):
         self.add(msg, msg_args=msg_args, level="transfer")
