@@ -1,8 +1,8 @@
 # COPYRIGHT (C) 2020-2022 Nicotine+ Contributors
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2016 Mutnick <muhing@yahoo.com>
-# COPYRIGHT (C) 2009-2011 Quinox <quinox@users.sf.net>
-# COPYRIGHT (C) 2009 Daelstorm <daelstorm@gmail.com>
+# COPYRIGHT (C) 2009-2011 quinox <quinox@users.sf.net>
+# COPYRIGHT (C) 2009 daelstorm <daelstorm@gmail.com>
 #
 # GNU GENERAL PUBLIC LICENSE
 #    Version 3, 29 June 2007
@@ -33,6 +33,7 @@ import time
 from pynicotine import slskmessages
 from pynicotine.logfacility import log
 from pynicotine.slskmessages import UINT_LIMIT
+from pynicotine.slskmessages import UserStatus
 from pynicotine.utils import TRANSLATE_PUNCTUATION
 from pynicotine.utils import encode_path
 from pynicotine.utils import rename_process
@@ -49,15 +50,22 @@ if importlib.util.find_spec("_gdbm"):
 
 elif importlib.util.find_spec("semidbm"):
 
+    import semidbm  # pylint: disable=import-error
+    try:
+        # semidbm throws an exception when calling sync on a read-only dict, avoid this
+        del semidbm.db._SemiDBMReadOnly.sync  # pylint: disable=protected-access
+
+    except AttributeError:
+        pass
+
     def shelve_open_semidbm(filename, flag='c', protocol=None, writeback=False):
-        import semidbm  # pylint: disable=import-error
         return shelve.Shelf(semidbm.open(filename, flag), protocol, writeback)
 
     shelve.open = shelve_open_semidbm
 
 else:
     log.add(_("Cannot find %(option1)s or %(option2)s, please install either one.") % {
-        "option1": "gdbm",
+        "option1": "python3-gdbm",
         "option2": "semidbm"
     })
     sys.exit()
@@ -88,13 +96,15 @@ class Scanner:
             from pynicotine.metadata.tinytag import TinyTag
             self.tinytag = TinyTag()
 
-            Shares.load_shares(self.share_dbs, self.share_db_paths)
+            if not Shares.load_shares(self.share_dbs, self.share_db_paths, remove_failed=True):
+                # Failed to load shares, rebuild
+                self.rescan = self.rebuild = True
 
             if self.init:
                 self.create_compressed_shares()
 
             if self.rescan:
-                start_num_folders = len(list(self.share_dbs["buddyfiles"]))
+                start_num_folders = len(list(self.share_dbs.get("buddyfiles", {})))
 
                 self.queue.put((_("Rescanning shares…"), None, None))
                 self.queue.put((_("%(num)s folders found before rescan, rebuilding…"),
@@ -143,6 +153,30 @@ class Scanner:
         self.create_compressed_shares_message("normal")
         self.create_compressed_shares_message("buddy")
 
+    def create_db_file(self, destination):
+
+        share_db = self.share_dbs.get(destination)
+
+        if share_db is not None:
+            share_db.close()
+
+        db_path = os.path.join(self.config.data_dir, destination + ".db")
+        self.remove_db_file(db_path)
+
+        return shelve.open(db_path, flag='n', protocol=pickle.HIGHEST_PROTOCOL)
+
+    @staticmethod
+    def remove_db_file(db_path):
+
+        db_path_encoded = encode_path(db_path)
+
+        if os.path.isfile(db_path_encoded):
+            os.remove(db_path_encoded)
+
+        elif os.path.isdir(db_path_encoded):
+            import shutil
+            shutil.rmtree(db_path_encoded)
+
     def set_shares(self, share_type, files=None, streams=None, mtimes=None, wordindex=None):
 
         self.config.create_data_folder()
@@ -162,12 +196,8 @@ class Scanner:
                 if share_type == "buddy":
                     destination = "buddy" + destination
 
-                # Close old db
-                self.share_dbs[destination].close()
-
-                self.share_dbs[destination] = shelve.open(os.path.join(self.config.data_dir, destination + ".db"),
-                                                          flag='n', protocol=pickle.HIGHEST_PROTOCOL)
-                self.share_dbs[destination].update(source)
+                self.share_dbs[destination] = share_db = self.create_db_file(destination)
+                share_db.update(source)
 
             except Exception as error:
                 self.queue.put((_("Can't save %(filename)s: %(error)s"),
@@ -203,8 +233,11 @@ class Scanner:
         if mtimes is not None:
             new_mtimes = {**mtimes, **new_mtimes}
 
-        old_mtimes = self.share_dbs[prefix + "mtimes"]
-        share_version = old_mtimes.get("__NICOTINE_SHARE_VERSION__")
+        old_mtimes = self.share_dbs.get(prefix + "mtimes")
+        share_version = None
+
+        if old_mtimes is not None:
+            share_version = old_mtimes.get("__NICOTINE_SHARE_VERSION__")
 
         # Rebuild shares if share version is outdated
         if share_version is None or share_version < self.version:
@@ -214,8 +247,8 @@ class Scanner:
             # Get mtimes for top-level shared folders, then every subfolder
             try:
                 files, streams, mtimes = self.get_files_list(
-                    folder, old_mtimes, self.share_dbs[prefix + "files"],
-                    self.share_dbs[prefix + "streams"], rebuild
+                    folder, old_mtimes, self.share_dbs.get(prefix + "files"),
+                    self.share_dbs.get(prefix + "streams"), rebuild
                 )
                 new_files = {**new_files, **files}
                 new_streams = {**new_streams, **streams}
@@ -351,7 +384,6 @@ class Scanner:
     def get_file_info(self, name, pathname, tinytag, file_stat=None):
         """ Get file metadata """
 
-        size = 0
         audio = None
         bitrate_info = None
         duration_info = None
@@ -407,9 +439,8 @@ class Scanner:
         For the word index db, we can't use the same approach, as we need to access
         dict elements frequently. This would take too long to access from disk. """
 
-        self.share_dbs[fileindex_dest].close()
-        self.share_dbs[fileindex_dest] = shelve.open(os.path.join(self.config.data_dir, fileindex_dest + ".db"),
-                                                     flag='n', protocol=pickle.HIGHEST_PROTOCOL)
+        self.share_dbs[fileindex_dest] = fileindex_db = self.create_db_file(fileindex_dest)
+
         wordindex = {}
         file_index = -1
         num_shared_files = len(shared_files)
@@ -429,7 +460,7 @@ class Scanner:
 
                 # Add to file index
                 fileinfo[0] = folder + '\\' + filename
-                self.share_dbs[fileindex_dest][repr(file_index)] = fileinfo
+                fileindex_db[repr(file_index)] = fileinfo
 
                 # Collect words from filenames for Search index
                 # Use set to prevent duplicates
@@ -524,7 +555,7 @@ class Shares:
                 return real
 
             if path.startswith(virtual + os.sep):
-                realpath = real + path[len(virtual):]
+                realpath = real.rstrip('/\\') + path[len(virtual):]
                 return realpath
 
         return "__INTERNAL_ERROR__" + path
@@ -540,6 +571,10 @@ class Shares:
     @staticmethod
     def get_normalized_virtual_name(virtual_name, shared_folders):
 
+        # Provide a default name for root folders
+        if not virtual_name:
+            virtual_name = "Shared"
+
         # Remove slashes from share name to avoid path conflicts
         virtual_name = virtual_name.replace('/', '_').replace('\\', '_')
         new_virtual_name = str(virtual_name)
@@ -562,7 +597,7 @@ class Shares:
             virtual = shared_folder.replace('/', '_').replace('\\', '_').strip('_')
             log.add("Renaming shared folder '%s' to '%s'. A rescan of your share is required.",
                     (shared_folder, virtual))
-            return (virtual, shared_folder)
+            return virtual, shared_folder
 
         self.config.sections["transfers"]["shared"] = [_convert_to_virtual(x)
                                                        for x in self.config.sections["transfers"]["shared"]]
@@ -570,49 +605,35 @@ class Shares:
                                                             for x in self.config.sections["transfers"]["buddyshared"]]
 
     @classmethod
-    def load_shares(cls, shares, dbs, reset_shares=False):
+    def load_shares(cls, shares, dbs, remove_failed=False):
 
         errors = []
         exception = None
 
-        for destination, shelvefile in dbs:
-            shelvefile_encoded = encode_path(shelvefile)
+        for destination, db_path in dbs:
+            db_path_encoded = encode_path(db_path)
 
             try:
-                if not reset_shares:
-                    shares[destination] = shelve.open(shelvefile, protocol=pickle.HIGHEST_PROTOCOL)
-                else:
-                    try:
-                        os.remove(shelvefile_encoded)
-
-                    except IsADirectoryError:
-                        # Potentially trying to use gdbm with a semidbm database
-                        os.rmdir(shelvefile_encoded)
-
-                    shares[destination] = shelve.open(shelvefile, flag='n', protocol=pickle.HIGHEST_PROTOCOL)
+                if os.path.exists(db_path_encoded):
+                    shares[destination] = shelve.open(db_path, flag='r', protocol=pickle.HIGHEST_PROTOCOL)
 
             except Exception:
                 from traceback import format_exc
 
-                shares[destination] = None
-                errors.append(shelvefile)
+                errors.append(db_path)
                 exception = format_exc()
 
+                if remove_failed:
+                    Scanner.remove_db_file(db_path)
+
         if not errors:
-            return
+            return True
 
         log.add(_("Failed to process the following databases: %(names)s"), {
             'names': '\n'.join(errors)
         })
         log.add(exception)
-
-        if not reset_shares:
-            log.add(_("Attempting to reset index of shared files due to an error. Please rescan your shares."))
-            cls.load_shares(shares, dbs, reset_shares=True)
-            return
-
-        log.add(_("File index of shared files could not be accessed. This could occur due to several instances of "
-                  "Nicotine+ being active simultaneously, file permission issues, or another issue in Nicotine+."))
+        return False
 
     def file_is_shared(self, user, virtualfilename, realfilename):
 
@@ -692,7 +713,10 @@ class Shares:
     def send_num_shared_folders_files(self):
         """ Send number publicly shared files to the server. """
 
-        if not (self.core and self.core.logged_in):
+        if not (self.core and self.core.user_status != UserStatus.OFFLINE):
+            return
+
+        if self.rescanning:
             return
 
         try:
@@ -700,7 +724,8 @@ class Shares:
             index = self.share_dbs.get("fileindex")
 
             if shared is None or index is None:
-                return
+                shared = {}
+                index = []
 
             try:
                 sharedfolders = len(shared)
@@ -751,7 +776,7 @@ class Shares:
         shared_folders = self.config.sections["transfers"]["shared"][:]
         shared_folders_buddy = self.config.sections["transfers"]["buddyshared"][:]
 
-        return (shared_folders, shared_folders_buddy)
+        return shared_folders, shared_folders_buddy
 
     def process_scanner_messages(self, scanner, scanner_queue):
 
@@ -823,12 +848,11 @@ class Shares:
 
         # Scanning done, load shares in the main process again
         self.load_shares(self.share_dbs, self.share_db_paths)
+        self.rescanning = False
 
         if not error:
             self.send_num_shared_folders_files()
 
-        self.rescanning = False
-
         # Process any file transfer queue requests that arrived while scanning
         if self.network_callback:
             self.network_callback(self.pending_network_msgs)
