@@ -1,10 +1,10 @@
 # COPYRIGHT (C) 2020-2022 Nicotine+ Contributors
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2016 Mutnick <muhing@yahoo.com>
-# COPYRIGHT (C) 2013 eL_vErDe <gandalf@le-vert.net>
-# COPYRIGHT (C) 2008-2012 Quinox <quinox@users.sf.net>
-# COPYRIGHT (C) 2009 Hedonist <ak@sensi.org>
-# COPYRIGHT (C) 2006-2009 Daelstorm <daelstorm@gmail.com>
+# COPYRIGHT (C) 2013 eLvErDe <gandalf@le-vert.net>
+# COPYRIGHT (C) 2008-2012 quinox <quinox@users.sf.net>
+# COPYRIGHT (C) 2009 hedonist <ak@sensi.org>
+# COPYRIGHT (C) 2006-2009 daelstorm <daelstorm@gmail.com>
 # COPYRIGHT (C) 2003-2004 Hyriand <hyriand@thegraveyard.org>
 # COPYRIGHT (C) 2001-2003 Alexander Kanavin
 #
@@ -256,11 +256,12 @@ class Transfers:
             transfer_list = self.downloads
 
         for transfer_row in transfers:
-            # User / filename / path
+            num_attributes = len(transfer_row)
 
-            if len(transfer_row) < 3:
+            if num_attributes < 3:
                 continue
 
+            # User / filename / path
             user = transfer_row[0]
 
             if not isinstance(user, str):
@@ -277,11 +278,10 @@ class Transfers:
                 continue
 
             # Status
+            loaded_status = None
 
-            try:
+            if num_attributes >= 4:
                 loaded_status = str(transfer_row[3])
-            except IndexError:
-                loaded_status = None
 
             if transfer_type == "uploads" and loaded_status != "Finished":
                 # Only finished uploads are supposed to be restored
@@ -297,36 +297,35 @@ class Transfers:
                 status = "User logged off"
 
             # Size / offset
+            size = 0
+            current_byte_offset = None
 
-            try:
-                size = int(transfer_row[4])
-            except Exception:
-                size = 0
+            if num_attributes >= 5:
+                loaded_size = transfer_row[4]
 
-            try:
-                current_byte_offset = int(transfer_row[5])
-            except Exception:
-                current_byte_offset = None
+                if loaded_size and isinstance(loaded_size, (int, float)):
+                    size = int(loaded_size)
+
+            if num_attributes >= 6:
+                loaded_byte_offset = transfer_row[5]
+
+                if loaded_byte_offset and isinstance(loaded_byte_offset, (int, float)):
+                    current_byte_offset = int(loaded_byte_offset)
 
             # Bitrate / length
-
             bitrate = length = None
 
-            try:
+            if num_attributes >= 7:
                 loaded_bitrate = transfer_row[6]
 
                 if loaded_bitrate is not None:
                     bitrate = str(loaded_bitrate)
-            except IndexError:
-                pass
 
-            try:
+            if num_attributes >= 8:
                 loaded_length = transfer_row[7]
 
                 if loaded_length is not None:
                     length = str(loaded_length)
-            except IndexError:
-                pass
 
             transfer_list.appendleft(
                 Transfer(
@@ -579,8 +578,7 @@ class Transfers:
         user_offline = (msg.status == UserStatus.OFFLINE)
         download_statuses = ("Queued", "Getting status", "Too many files", "Too many megabytes", "Pending shutdown.",
                              "User logged off", "Connection timeout", "Remote file error", "Cancelled")
-        upload_statuses = ("Getting status", "Disallowed extension", "User logged off", "Connection timeout",
-                           "Cancelled")
+        upload_statuses = ("Getting status", "User logged off", "Connection timeout")
 
         for download in reversed(self.downloads.copy()):
             if (download.user == username
@@ -616,35 +614,62 @@ class Transfers:
         if self.uploadsview and update:
             self.uploadsview.update_model()
 
-    def get_cant_connect_queue_file(self, username, filename):
+    def get_cant_connect_queue_file(self, username, filename, offline):
         """ We can't connect to the user, either way (QueueUpload). """
 
         for download in self.downloads:
             if download.filename != filename or download.user != username:
                 continue
 
-            download.status = "Connection timeout"
+            log.add_transfer("Download attempt for file %(filename)s from user %(user)s timed out", {
+                "filename": filename,
+                "user": username
+            })
+
+            download.status = "User logged off" if offline else "Connection timeout"
             download.token = None
 
+            if download in self.transfer_request_times:
+                del self.transfer_request_times[download]
+
             self.update_download(download)
             self.core.watch_user(username)
             break
 
-    def get_cant_connect_upload(self, token):
+    def get_cant_connect_upload(self, username, token, offline):
         """ We can't connect to the user, either way (TransferRequest, FileUploadInit). """
 
         for upload in self.uploads:
-            if upload.token != token:
+            if upload.token != token or upload.user != username:
                 continue
 
-            upload.status = "Connection timeout"
+            log.add_transfer("Upload attempt for file %(filename)s with token %(token)s to user %(user)s timed out", {
+                "filename": upload.filename,
+                "token": token,
+                "user": username
+            })
+
+            if upload.sock is not None:
+                log.add_transfer("Existing file connection for upload with token %s already exists?", token)
+                return
+
+            if offline:
+                upload.status = "User logged off"
+                upload_cleared = self.auto_clear_upload(upload)
+            else:
+                upload.status = "Connection timeout"
+                upload_cleared = False
+
             upload.token = None
             upload.queue_position = 0
 
-            self.update_user_counter(upload.user)
-            self.update_upload(upload)
+            if upload in self.transfer_request_times:
+                del self.transfer_request_times[upload]
 
-            self.core.watch_user(upload.user)
+            if not upload_cleared:
+                self.update_upload(upload)
+
+            self.core.watch_user(username)
             self.check_upload_queue()
             return
 
@@ -689,9 +714,6 @@ class Transfers:
                         username, virtualpath, path=destination,
                         size=size, bitrate=h_bitrate, length=h_length)
 
-                if directory in self.requested_folders.get(username, []):
-                    del self.requested_folders[username][directory]
-
     def queue_upload(self, msg):
         """ Peer code: 43 """
         """ Peer remotely queued a download (upload here). This is the modern replacement to
@@ -709,7 +731,7 @@ class Transfers:
         real_path = self.core.shares.virtual2real(filename)
         allowed, reason = self.check_queue_upload_allowed(user, msg.init.addr, filename, real_path, msg)
 
-        log.add_transfer(("Upload request for file %(filename)s from user %(user)s: "
+        log.add_transfer(("Upload request for file %(filename)s from user: %(user)s, "
                           "allowed: %(allowed)s, reason: %(reason)s"), {
             "filename": filename,
             "user": user,
@@ -735,13 +757,12 @@ class Transfers:
         """ Peer code: 40 """
 
         user = msg.init.target_user
-        response = None
 
         if msg.direction == TransferDirection.UPLOAD:
             response = self.transfer_request_downloads(msg)
 
-            log.add_transfer(("Responding to download request %(token)s for file %(filename)s "
-                              "from user %(user)s: allowed: %(allowed)s, reason: %(reason)s"), {
+            log.add_transfer(("Responding to download request with token %(token)s for file %(filename)s "
+                              "from user: %(user)s, allowed: %(allowed)s, reason: %(reason)s"), {
                 "token": response.token, "filename": msg.file, "user": user,
                 "allowed": response.allowed, "reason": response.reason
             })
@@ -753,7 +774,7 @@ class Transfers:
                 return
 
             log.add_transfer(("Responding to legacy upload request %(token)s for file %(filename)s "
-                              "from user %(user)s: allowed: %(allowed)s, reason: %(reason)s"), {
+                              "from user %(user)s, allowed: %(allowed)s, reason: %(reason)s"), {
                 "token": response.token, "filename": msg.file, "user": user,
                 "allowed": response.allowed, "reason": response.reason
             })
@@ -774,7 +795,7 @@ class Transfers:
         size = msg.filesize
         token = msg.token
 
-        log.add_transfer("Received download request %(token)s for file %(filename)s from user %(user)s", {
+        log.add_transfer("Received download request with token %(token)s for file %(filename)s from user %(user)s", {
             "token": token,
             "filename": filename,
             "user": user
@@ -834,19 +855,19 @@ class Transfers:
                 path = os.path.join(self.config.sections["transfers"]["uploaddir"], user, parentdir)
 
             transfer = Transfer(user=user, filename=filename, path=path, status="Queued",
-                                size=size, token=msg.token)
+                                size=size, token=token)
             self.downloads.appendleft(transfer)
             self.update_download(transfer)
             self.core.watch_user(user)
 
-            return slskmessages.TransferResponse(allowed=False, reason="Queued", token=transfer.token)
+            return slskmessages.TransferResponse(allowed=False, reason="Queued", token=token)
 
         log.add_transfer("Denied file request: User %(user)s, %(msg)s", {
             'user': user,
             'msg': str(vars(msg))
         })
 
-        return slskmessages.TransferResponse(allowed=False, reason=cancel_reason, token=msg.token)
+        return slskmessages.TransferResponse(allowed=False, reason=cancel_reason, token=token)
 
     def transfer_request_uploads(self, msg):
         """ Remote peer is requesting to download a file through your upload queue.
@@ -855,9 +876,10 @@ class Transfers:
 
         user = msg.init.target_user
         filename = msg.file
+        token = msg.token
 
         log.add_transfer("Received legacy upload request %(token)s for file %(filename)s from user %(user)s", {
-            "token": msg.token,
+            "token": token,
             "filename": filename,
             "user": user
         })
@@ -868,7 +890,7 @@ class Transfers:
 
         if not allowed:
             if reason:
-                return slskmessages.TransferResponse(allowed=False, reason=reason, token=msg.token)
+                return slskmessages.TransferResponse(allowed=False, reason=reason, token=token)
 
             return None
 
@@ -892,27 +914,28 @@ class Transfers:
             self.append_upload(user, filename, transfer)
             self.update_upload(transfer)
 
-            return slskmessages.TransferResponse(allowed=False, reason="Queued", token=msg.token)
+            return slskmessages.TransferResponse(allowed=False, reason="Queued", token=token)
 
         # All checks passed, starting a new upload.
         size = self.get_file_size(real_path)
         transfer = Transfer(user=user, filename=filename, path=os.path.dirname(real_path),
-                            status="Getting status", token=msg.token, size=size)
+                            status="Getting status", token=token, size=size)
 
         self.transfer_request_times[transfer] = time.time()
         self.append_upload(user, filename, transfer)
         self.update_upload(transfer)
 
-        return slskmessages.TransferResponse(allowed=True, token=msg.token, filesize=size)
+        return slskmessages.TransferResponse(allowed=True, token=token, filesize=size)
 
     def transfer_response(self, msg):
         """ Peer code: 41 """
         """ Received a response to the file request from the peer """
 
+        username = msg.init.target_user
         token = msg.token
         reason = msg.reason
 
-        log.add_transfer(("Received response for upload request %(token)s: allowed: %(allowed)s, "
+        log.add_transfer(("Received response for upload with token: %(token)s, allowed: %(allowed)s, "
                           "reason: %(reason)s, file size: %(size)s"), {
             "token": token,
             "allowed": msg.allowed,
@@ -926,9 +949,13 @@ class Transfers:
                 reason = "Cancelled"
 
             for upload in self.uploads:
-                if upload.token != token:
+                if upload.token != token or upload.user != username:
                     continue
 
+                if upload.sock is not None:
+                    log.add_transfer("Upload with token %s already has an existing file connection", token)
+                    return
+
                 upload.status = reason
                 upload.token = None
                 upload.queue_position = 0
@@ -938,8 +965,6 @@ class Transfers:
                 if upload in self.transfer_request_times:
                     del self.transfer_request_times[upload]
 
-                self.update_user_counter(upload.user)
-
                 if reason in ("Complete", "Finished"):
                     # A complete download of this file already exists on the user's end
                     self.upload_finished(upload)
@@ -953,9 +978,13 @@ class Transfers:
             return
 
         for upload in self.uploads:
-            if upload.token != token:
+            if upload.token != token or upload.user != username:
                 continue
 
+            if upload.sock is not None:
+                log.add_transfer("Upload with token %s already has an existing file connection", token)
+                return
+
             self.core.send_message_to_peer(upload.user, slskmessages.FileUploadInit(None, token=token))
             self.check_upload_queue()
             return
@@ -966,10 +995,6 @@ class Transfers:
 
         transfer = msg.transfer
 
-        if transfer.status == "Transferring":
-            # Check if the transfer has started since the timeout callback was initiated
-            return
-
         log.add_transfer("Transfer %(filename)s with token %(token)s for user %(user)s timed out", {
             "filename": transfer.filename,
             "token": transfer.token,
@@ -977,7 +1002,7 @@ class Transfers:
         })
 
         transfer.status = "Connection timeout"
-        transfer.token = None
+        self.abort_transfer(transfer)
 
         self.core.watch_user(transfer.user)
 
@@ -986,7 +1011,6 @@ class Transfers:
 
         elif transfer in self.uploads:
             transfer.queue_position = 0
-            self.update_user_counter(transfer.user)
             self.update_upload(transfer)
 
         if transfer in self.transfer_request_times:
@@ -997,14 +1021,15 @@ class Transfers:
     def download_file_error(self, msg):
         """ Networking thread encountered a local file error for download """
 
-        sock = msg.sock
+        username = msg.user
+        token = msg.token
 
         for download in self.downloads:
-            if download.sock != sock:
+            if download.token != token or download.user != username:
                 continue
 
-            self.abort_transfer(download)
             download.status = "Local file error"
+            self.abort_transfer(download)
 
             log.add(_("Download I/O error: %s"), msg.error)
             self.update_download(download)
@@ -1013,14 +1038,15 @@ class Transfers:
     def upload_file_error(self, msg):
         """ Networking thread encountered a local file error for upload """
 
-        sock = msg.sock
+        username = msg.user
+        token = msg.token
 
         for upload in self.uploads:
-            if upload.sock != sock:
+            if upload.token != token or upload.user != username:
                 continue
 
-            self.abort_transfer(upload)
             upload.status = "Local file error"
+            self.abort_transfer(upload)
 
             log.add(_("Upload I/O error: %s"), msg.error)
             self.update_upload(upload)
@@ -1030,13 +1056,13 @@ class Transfers:
     def file_download_init(self, msg):
         """ A peer is requesting to start uploading a file to us """
 
+        username = msg.init.target_user
         token = msg.token
 
         for download in self.downloads:
-            if download.token != token:
+            if download.token != token or download.user != username:
                 continue
 
-            username = download.user
             filename = download.filename
 
             log.add_transfer(("Received file download init with token %(token)s for file %(filename)s "
@@ -1048,16 +1074,12 @@ class Transfers:
 
             if download.sock is not None:
                 log.add_transfer("Download already has an existing file connection, ignoring init message")
-                self.queue.append(slskmessages.ConnClose(msg.init.sock))
+                self.queue.append(slskmessages.CloseConnection(msg.init.sock))
                 return
 
             incomplete_folder = self.config.sections["transfers"]["incompletedir"]
             need_update = True
             download.sock = msg.init.sock
-            download.token = None
-
-            if download in self.transfer_request_times:
-                del self.transfer_request_times[download]
 
             if not incomplete_folder:
                 if download.path:
@@ -1080,7 +1102,6 @@ class Transfers:
                 self.download_folder_error(download, error)
 
             else:
-                file_handle = None
                 try:
                     incomplete_path = self.get_incomplete_file_path(incomplete_folder, username, filename)
                     file_handle = open(encode_path(incomplete_path), 'ab+')  # pylint: disable=consider-using-with
@@ -1106,8 +1127,8 @@ class Transfers:
                 except OSError as error:
                     log.add(_("Download I/O error: %s"), error)
 
-                    self.abort_transfer(download)
                     download.status = "Local file error"
+                    self.abort_transfer(download)
 
                 else:
                     download.file = file_handle
@@ -1129,7 +1150,7 @@ class Transfers:
                     if download.size > offset:
                         download.status = "Transferring"
                         self.queue.append(slskmessages.DownloadFile(
-                            sock=download.sock, file=file_handle, leftbytes=(download.size - offset)
+                            init=msg.init, token=token, file=file_handle, leftbytes=(download.size - offset)
                         ))
                         self.queue.append(slskmessages.FileOffset(init=msg.init, offset=offset))
 
@@ -1146,6 +1167,10 @@ class Transfers:
             return
 
         # Support legacy transfer system (clients: old Nicotine+ versions, slskd)
+        # The user who requested the download initiates the file upload connection
+        # in this case, but we always assume an incoming file init message is
+        # FileDownloadInit
+
         log.add_transfer(("Received unknown file download init message with token %s, checking if peer "
                           "requested us to upload a file instead"), token)
         self.file_upload_init(msg)
@@ -1153,13 +1178,13 @@ class Transfers:
     def file_upload_init(self, msg):
         """ We are requesting to start uploading a file to a peer """
 
+        username = msg.init.target_user
         token = msg.token
 
         for upload in self.uploads:
-            if upload.token != token:
+            if upload.token != token or upload.user != username:
                 continue
 
-            username = upload.user
             filename = upload.filename
 
             log.add_transfer("Initializing upload with token %(token)s for file %(filename)s to user %(user)s", {
@@ -1170,16 +1195,11 @@ class Transfers:
 
             if upload.sock is not None:
                 log.add_transfer("Upload already has an existing file connection, ignoring init message")
-                self.queue.append(slskmessages.ConnClose(msg.init.sock))
+                self.queue.append(slskmessages.CloseConnection(msg.init.sock))
                 return
 
             need_update = True
             upload.sock = msg.init.sock
-            upload.token = None
-            file_handle = None
-
-            if upload in self.transfer_request_times:
-                del self.transfer_request_times[upload]
 
             real_path = self.core.shares.virtual2real(filename)
 
@@ -1221,7 +1241,9 @@ class Transfers:
 
                 if upload.size > 0:
                     upload.status = "Transferring"
-                    self.queue.append(slskmessages.UploadFile(upload.sock, file=file_handle, size=upload.size))
+                    self.queue.append(slskmessages.UploadFile(
+                        init=msg.init, token=token, file=file_handle, size=upload.size
+                    ))
 
                 else:
                     self.upload_finished(upload, file_handle=file_handle)
@@ -1236,7 +1258,7 @@ class Transfers:
             return
 
         log.add_transfer("Unknown file upload init message with token %s", token)
-        self.queue.append(slskmessages.ConnClose(msg.init.sock))
+        self.queue.append(slskmessages.CloseConnection(msg.init.sock))
 
     def upload_denied(self, msg):
         """ Peer code: 50 """
@@ -1325,10 +1347,11 @@ class Transfers:
     def file_download(self, msg):
         """ A file download is in progress """
 
-        sock = msg.sock
+        username = msg.init.target_user
+        token = msg.token
 
         for download in self.downloads:
-            if download.sock != sock:
+            if download.token != token or download.user != username:
                 continue
 
             if download in self.transfer_request_times:
@@ -1360,10 +1383,11 @@ class Transfers:
     def file_upload(self, msg):
         """ A file upload is in progress """
 
-        sock = msg.sock
+        username = msg.init.target_user
+        token = msg.token
 
         for upload in self.uploads:
-            if upload.sock != sock:
+            if upload.token != token or upload.user != username:
                 continue
 
             if upload in self.transfer_request_times:
@@ -1395,39 +1419,40 @@ class Transfers:
             self.update_upload(upload)
             return
 
-    def download_conn_close(self, msg):
-        """ The remote peer has closed a file transfer connection """
+    def download_connection_closed(self, msg):
+        """ A file download connection has closed for any reason """
 
-        sock = msg.sock
+        username = msg.user
+        token = msg.token
 
         for download in self.downloads:
-            if download.sock != sock:
+            if download.token != token or download.user != username:
                 continue
 
             if download.current_byte_offset is not None and download.current_byte_offset >= download.size:
                 self.download_finished(download, file_handle=download.file)
                 return
 
-            self.abort_transfer(download)
-
             if download.status != "Finished":
-                if self.user_logged_out(download.user):
+                if self.core.user_statuses.get(download.user) == UserStatus.OFFLINE:
                     download.status = "User logged off"
                 else:
                     download.status = "Cancelled"
 
+            self.abort_transfer(download)
             self.update_download(download)
             return
 
-    def upload_conn_close(self, msg):
-        """ The remote peer has closed a file transfer connection """
+    def upload_connection_closed(self, msg):
+        """ A file upload connection has closed for any reason """
 
-        sock = msg.sock
+        username = msg.user
+        token = msg.token
         timed_out = msg.timed_out
 
         # We need a copy due to upload auto-clearing modifying the deque during iteration
         for upload in self.uploads.copy():
-            if upload.sock != sock:
+            if upload.token != token or upload.user != username:
                 continue
 
             if not timed_out and upload.current_byte_offset is not None and upload.current_byte_offset >= upload.size:
@@ -1445,9 +1470,7 @@ class Transfers:
             if upload.status == "Finished":
                 return
 
-            self.abort_transfer(upload)
-
-            if self.user_logged_out(upload.user):
+            if self.core.user_statuses.get(upload.user) == UserStatus.OFFLINE:
                 upload.status = "User logged off"
             else:
                 upload.status = "Cancelled"
@@ -1456,6 +1479,8 @@ class Transfers:
                 # intentionally cancelled, the peer should ignore this message.
                 self.core.send_message_to_peer(upload.user, slskmessages.UploadFailed(file=upload.filename))
 
+            self.abort_transfer(upload)
+
             if not self.auto_clear_upload(upload):
                 self.update_upload(upload)
 
@@ -1529,7 +1554,7 @@ class Transfers:
     """ Transfer Actions """
 
     def get_folder(self, user, folder):
-        self.core.send_message_to_peer(user, slskmessages.FolderContentsRequest(directory=folder))
+        self.core.send_message_to_peer(user, slskmessages.FolderContentsRequest(directory=folder, token=1))
 
     def get_file(self, user, filename, path="", transfer=None, size=0, bitrate=None, length=None, ui_callback=True):
 
@@ -1556,6 +1581,7 @@ class Transfers:
         else:
             transfer.filename = filename
             transfer.status = "Queued"
+            transfer.token = None
 
         self.core.watch_user(user)
 
@@ -1565,9 +1591,9 @@ class Transfers:
 
                 if downloadregexp.search(filename) is not None:
                     log.add_transfer("Filtering: %s", filename)
-                    self.abort_transfer(transfer)
-                    # The string to be displayed on the GUI
+
                     transfer.status = "Filtered"
+                    self.abort_transfer(transfer)
 
                     if self.auto_clear_download(transfer):
                         return
@@ -1575,7 +1601,8 @@ class Transfers:
             except re.error:
                 pass
 
-        if self.user_logged_out(user):
+        if UserStatus.OFFLINE in (self.core.user_status, self.core.user_statuses.get(user)):
+            # Either we are offline or the user we want to download from is
             transfer.status = "User logged off"
 
         elif transfer.status != "Filtered":
@@ -1600,9 +1627,6 @@ class Transfers:
 
     def push_file(self, user, filename, size, path="", transfer=None, bitrate=None, length=None, locally_queued=False):
 
-        if not self.core.logged_in:
-            return
-
         real_path = self.core.shares.virtual2real(filename)
         size_attempt = self.get_file_size(real_path)
 
@@ -1623,6 +1647,7 @@ class Transfers:
             transfer.filename = filename
             transfer.size = size
             transfer.status = "Queued"
+            transfer.token = None
 
         log.add_transfer("Initializing upload request for file %(file)s to user %(user)s", {
             'file': filename,
@@ -1631,7 +1656,8 @@ class Transfers:
 
         self.core.watch_user(user)
 
-        if self.user_logged_out(user):
+        if UserStatus.OFFLINE in (self.core.user_status, self.core.user_statuses.get(user)):
+            # Either we are offline or the user we want to upload to is
             transfer.status = "User logged off"
 
             if not self.auto_clear_upload(transfer):
@@ -1644,7 +1670,7 @@ class Transfers:
             transfer.status = "Getting status"
             self.transfer_request_times[transfer] = time.time()
 
-            log.add_transfer(("Requesting to upload file %(filename)s with token %(token)s to user %(user)s"), {
+            log.add_transfer("Requesting to upload file %(filename)s with token %(token)s to user %(user)s", {
                 "filename": filename,
                 "token": transfer.token,
                 "user": user
@@ -1695,9 +1721,6 @@ class Transfers:
             self.uploads.insert(old_index, transferobj)
             return
 
-        if user not in self.user_update_counters:
-            self.update_user_counter(user)
-
         self.uploads.appendleft(transferobj)
 
     def can_upload(self, user):
@@ -1727,29 +1750,21 @@ class Transfers:
 
         return False
 
-    def user_logged_out(self, user):
-        """ Check if a user who previously queued a file has logged out since """
+    def get_folder_destination(self, user, folder, remove_prefix="", remove_destination=True):
 
-        if not self.core.logged_in:
-            return True
+        if not remove_prefix and '\\' in folder:
+            remove_prefix = folder.rsplit('\\', 1)[0]
 
-        if user not in self.core.user_statuses:
-            return False
-
-        return self.core.user_statuses[user] == UserStatus.OFFLINE
-
-    def get_folder_destination(self, user, directory, remove_prefix=""):
-
-        if not remove_prefix and '\\' in directory:
-            remove_prefix = directory.rsplit('\\', 1)[0]
-
-        # Get the last folders in directory path, excluding remove_prefix
-        target_folders = directory.replace(remove_prefix, "").lstrip('\\').replace('\\', os.sep)
+        # Get the last folders in folder path, excluding remove_prefix
+        target_folders = folder.replace(remove_prefix, "").lstrip('\\').replace('\\', os.sep)
 
         # Check if a custom download location was specified
-        if (user in self.requested_folders and directory in self.requested_folders[user]
-                and self.requested_folders[user][directory]):
-            download_location = self.requested_folders[user].pop(directory)
+        if (user in self.requested_folders and folder in self.requested_folders[user]
+                and self.requested_folders[user][folder]):
+            download_location = self.requested_folders[user][folder]
+
+            if remove_destination:
+                del self.requested_folders[user][folder]
         else:
             download_location = self.get_default_download_folder(user)
 
@@ -1926,8 +1941,8 @@ class Transfers:
 
     def download_folder_error(self, transfer, error):
 
-        self.abort_transfer(transfer)
         transfer.status = "Download folder error"
+        self.abort_transfer(transfer)
 
         self.core.notifications.new_text_notification(
             _("OS error: %s") % error, title=_("Download folder error"))
@@ -1962,6 +1977,7 @@ class Transfers:
         transfer.status = "Finished"
         transfer.current_byte_offset = transfer.size
         transfer.sock = None
+        transfer.token = None
 
         self.core.statistics.append_stat_value("completed_downloads", 1)
 
@@ -1993,8 +2009,7 @@ class Transfers:
         transfer.status = "Finished"
         transfer.current_byte_offset = transfer.size
         transfer.sock = None
-
-        self.update_user_counter(transfer.user)
+        transfer.token = None
 
         log.add_upload(
             _("Upload finished: user %(user)s, IP address %(ip)s, file %(file)s"), {
@@ -2030,6 +2045,7 @@ class Transfers:
     def auto_clear_upload(self, transfer):
 
         if self.config.sections["transfers"]["autoclear_uploads"]:
+            self.update_user_counter(transfer.user)
             self.uploads.remove(transfer)
 
             if self.uploadsview:
@@ -2046,9 +2062,23 @@ class Transfers:
 
     def update_upload(self, transfer, update_parent=True):
 
+        user = transfer.user
+        status = transfer.status
+
         if self.uploadsview:
             self.uploadsview.update_model(transfer, update_parent=update_parent)
 
+        if status == "Queued" and user in self.user_update_counters:
+            # Don't update existing user counter for queued uploads
+            # We don't want to push the user back in the queue if they enqueued new files
+            return
+
+        if status == "Transferring":
+            # Avoid unnecessary updates while transferring
+            return
+
+        self.update_user_counter(user)
+
     def _check_transfer_timeouts(self):
 
         while True:
@@ -2334,6 +2364,8 @@ class Transfers:
         for user in users:
             self.core.network_filter.ban_user(user)
 
+        self.check_upload_queue()
+
     def retry_download(self, transfer):
 
         if transfer.status in ("Transferring", "Finished"):
@@ -2369,13 +2401,21 @@ class Transfers:
 
     def abort_transfer(self, transfer, reason="Cancelled", send_fail_message=False):
 
+        log.add_transfer(("Aborting transfer, user \"%(user)s\", filename \"%(filename)s\", token \"%(token)s\", "
+                          "status \"%(status)s\""), {
+            "user": transfer.user,
+            "filename": transfer.filename,
+            "token": transfer.token,
+            "status": transfer.status
+        })
+
         transfer.legacy_attempt = False
         transfer.size_changed = False
         transfer.token = None
         transfer.queue_position = 0
 
         if transfer.sock is not None:
-            self.queue.append(slskmessages.ConnClose(transfer.sock))
+            self.queue.append(slskmessages.CloseConnection(transfer.sock))
             transfer.sock = None
 
         if transfer in self.transfer_request_times:
@@ -2385,8 +2425,6 @@ class Transfers:
             self.close_file(transfer.file, transfer)
 
             if transfer in self.uploads:
-                self.update_user_counter(transfer.user)
-                self.check_upload_queue()
                 log.add_upload(
                     _("Upload aborted, user %(user)s file %(file)s"), {
                         "user": transfer.user,
@@ -2464,8 +2502,8 @@ class Transfers:
 
         for download in self.downloads:
             if download.status not in ("Finished", "Filtered", "Paused"):
-                self.abort_transfer(download)
                 download.status = "User logged off"
+                self.abort_transfer(download)
                 need_update = True
 
         if self.downloadsview and need_update:
